@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2002-2023, Florian Kirsch,
+// Copyright (C) 2002-2025, Florian Kirsch,
 // Hasso-Plattner-Institute at the University of Potsdam, Germany
 //
 // This library is free software; you can redistribute it and/or modify
@@ -24,6 +24,7 @@
 
 #include <opencsg.h>
 #include "displaylistPrimitive.h"
+#include <algorithm>
 #include <iostream>
 #include <sstream>
 #include <string>
@@ -32,27 +33,60 @@
 // of exit() with Visual Studio 2010
 #include "includeGl.h"
 
-enum { 
-    CSG_BASIC, CSG_WIDGET, CSG_GRID2D, CSG_GRID3D, CSG_CUBERACK, CSG_CONCAVE,
+enum {
+    BENCH_START,
 
-    ALGO_AUTOMATIC, GF_STANDARD, GF_DC, GF_OQ, SCS_STANDARD, SCS_DC, SCS_OQ
+    CSG_BASIC, CSG_WIDGET, CSG_GRID2D, CSG_GRID3D, CSG_CUBERACK, CSG_PIPE, CSG_CONCAVE,
+
+    ALGO_AUTOMATIC, GF_STANDARD, GF_DC, GF_OQ, SCS_STANDARD, SCS_DC, SCS_OQ,
+
+    CAM_OUTSIDE_DEFAULT, CAM_OUTSIDE_PERF, CAM_OUTSIDE_NOPERF, CAM_INSIDE, CAM_INSIDE_DEFAULT,
+
+    GLSL_FIXED_FUNCTION, GLSL_FTRANSFORM, GLSL_MVP_COMBINED, GLSL_MVP_SEPARATE
 };
 
 std::vector<OpenCSG::Primitive*> primitives;
+std::vector<OpenCSG::Primitive*> primitives2;
+std::vector<OpenCSG::Primitive*> primitives3;
+
 
 bool               spin = true;
+bool               inside = false;
 float              rot = 0.0f;
 std::ostringstream fpsStream;
 
-void clearPrimitives() {
-    for (std::vector<OpenCSG::Primitive*>::const_iterator i = primitives.begin(); i != primitives.end(); ++i) {
-        OpenCSG::DisplayListPrimitive* p = 
-            static_cast<OpenCSG::DisplayListPrimitive*>(*i);
-        glDeleteLists(1, p->getDisplayListId());
-        delete p;
+std::vector<GLuint> displaylistGarbagePile;
+
+bool               benchmode = false;
+bool               benchSettingFrameOne = false;
+int                benchShape = BENCH_START;
+int                benchAlgorithm = GF_STANDARD;
+int                benchPerfOption = CAM_OUTSIDE_DEFAULT;
+
+void clearPrimitives(std::vector<OpenCSG::Primitive*> & p)
+{
+    for (std::vector<OpenCSG::Primitive*>::const_iterator it = p.begin(); it != p.end(); ++it)
+    {
+        OpenCSG::DisplayListPrimitive* primitive = static_cast<OpenCSG::DisplayListPrimitive*>(*it);
+        GLuint id = primitive->getDisplayListId();
+        if (std::find(displaylistGarbagePile.begin(), displaylistGarbagePile.end(), id) == displaylistGarbagePile.end())
+        {
+            glDeleteLists(1, id);
+            displaylistGarbagePile.push_back(id);
+        }
+        delete primitive;
     }
-    
-    primitives.clear();
+
+    p.clear();
+}
+
+void clearPrimitives()
+{
+    displaylistGarbagePile.clear();
+    clearPrimitives(primitives);
+    clearPrimitives(primitives2);
+    clearPrimitives(primitives3);
+    displaylistGarbagePile.clear();
 }
 
 void solidCylinder(GLdouble radius, GLdouble height, GLint slices, GLint stacks) {
@@ -104,7 +138,7 @@ void setBasicShape() {
 }
 
 void setWidget() {
-    
+
     clearPrimitives();
 
     GLuint id1 = glGenLists(1);
@@ -151,7 +185,7 @@ void setWidget() {
 }
 
 void setGrid2D() {
-    
+
     clearPrimitives();
 
     GLuint id1 = glGenLists(1);
@@ -181,7 +215,7 @@ void setGrid2D() {
 }
 
 void setGrid3D() {
-    
+
     clearPrimitives();
 
     GLuint id1 = glGenLists(1);
@@ -245,8 +279,81 @@ void setCubeRack() {
     }
 }
 
+void setPipe() {
+
+    clearPrimitives();
+
+    GLuint id1 = glGenLists(1);
+    glNewList(id1, GL_COMPILE);
+    glPushMatrix();
+    glTranslatef(0.0f, 0.0f, -2.5f);
+    solidCylinder(0.6, 2.5, 20, 20);
+    glPopMatrix();
+    glEndList();
+
+    GLuint id2 = glGenLists(1);
+    glNewList(id2, GL_COMPILE);
+    glPushMatrix();
+    glTranslatef(0.0f, 0.0f, -2.51f);
+    solidCylinder(0.5, 2.52, 20, 20);
+    glPopMatrix();
+    glEndList();
+
+    GLuint id3 = glGenLists(1);
+    glNewList(id3, GL_COMPILE);
+    glPushMatrix();
+    glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
+    glTranslatef(0.0f, 0.0f, -2.5f);
+    solidCylinder(0.6, 2.5, 20, 20);
+    glPopMatrix();
+    glEndList();
+
+    GLuint id4 = glGenLists(1);
+    glNewList(id4, GL_COMPILE);
+    glPushMatrix();
+    glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
+    glTranslatef(0.0f, 0.0f, -2.51f);
+    solidCylinder(0.5, 2.52, 20, 20);
+    glPopMatrix();
+    glEndList();
+
+    GLuint id5 = glGenLists(1);
+    glNewList(id5, GL_COMPILE);
+    glutSolidSphere(0.6, 20, 20);
+    glEndList();
+
+    GLuint id6 = glGenLists(1);
+    glNewList(id6, GL_COMPILE);
+    glutSolidSphere(0.5, 20, 20);
+    glEndList();
+
+    GLuint id7 = glGenLists(1);
+    glNewList(id7, GL_COMPILE);
+    glPushMatrix();
+    glTranslatef(-0.8f, 1.0f, -0.8f);
+    glutSolidCube(2.0);
+    glPopMatrix();
+    glEndList();
+
+    primitives.push_back(new OpenCSG::DisplayListPrimitive(id1, OpenCSG::Intersection, 1));
+    primitives.push_back(new OpenCSG::DisplayListPrimitive(id2, OpenCSG::Subtraction, 1));
+    primitives.push_back(new OpenCSG::DisplayListPrimitive(id4, OpenCSG::Subtraction, 1));
+    primitives.push_back(new OpenCSG::DisplayListPrimitive(id7, OpenCSG::Subtraction, 1));
+
+    primitives2.push_back(new OpenCSG::DisplayListPrimitive(id3, OpenCSG::Intersection, 1));
+    primitives2.push_back(new OpenCSG::DisplayListPrimitive(id2, OpenCSG::Subtraction, 1));
+    primitives2.push_back(new OpenCSG::DisplayListPrimitive(id4, OpenCSG::Subtraction, 1));
+    primitives2.push_back(new OpenCSG::DisplayListPrimitive(id7, OpenCSG::Subtraction, 1));
+
+    primitives3.push_back(new OpenCSG::DisplayListPrimitive(id5, OpenCSG::Intersection, 1));
+    primitives3.push_back(new OpenCSG::DisplayListPrimitive(id6, OpenCSG::Subtraction, 1));
+    primitives3.push_back(new OpenCSG::DisplayListPrimitive(id2, OpenCSG::Subtraction, 1));
+    primitives3.push_back(new OpenCSG::DisplayListPrimitive(id4, OpenCSG::Subtraction, 1));
+    primitives3.push_back(new OpenCSG::DisplayListPrimitive(id7, OpenCSG::Subtraction, 1));
+}
+
 void setConcave() {
-    
+
     clearPrimitives();
 
     GLuint id1 = glGenLists(1);
@@ -312,62 +419,63 @@ void display()
 {
     glMatrixMode(GL_MODELVIEW);
     glLoadIdentity();
-    gluLookAt(0.0, 2.0, 5.0,  /* eye is at (0,2,5) */
-              0.0, 0.0, 0.0,  /* center is at (0,0,0) */
-              0.0, 1.0, 0.0); /* up is in positive Y direction */
+
+    if (inside)
+    {
+       gluLookAt(0.0, 0.25, 0.0,  /* eye is slightly above (0,0,0) */
+                 1.0,-0.25, 0.0,  /* looking to the right and a bit down */
+                 0.0, 1.0, 0.0);  /* up is in positive Y direction */
+    }
+    else
+    {
+        gluLookAt(0.0, 2.0, 5.0,  /* eye is at (0,2,5) */
+                  0.0, 0.0, 0.0,  /* center is at (0,0,0) */
+                  0.0, 1.0, 0.0); /* up is in positive Y direction */
+    }
     glRotatef(rot, 0.0f, 1.0f, 0.0f);
 
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
     OpenCSG::render(primitives);
+    OpenCSG::render(primitives2);
+    OpenCSG::render(primitives3);
     glDepthFunc(GL_EQUAL);
-    for (std::vector<OpenCSG::Primitive*>::const_iterator i = primitives.begin(); i != primitives.end(); ++i) {
-        (*i)->render();
+    std::vector<OpenCSG::Primitive*>::const_iterator itr;
+    for (itr = primitives.begin(); itr != primitives.end(); ++itr) {
+        (*itr)->render();
+    }
+    for (itr = primitives2.begin(); itr != primitives2.end(); ++itr) {
+        (*itr)->render();
+    }
+    for (itr = primitives3.begin(); itr != primitives3.end(); ++itr) {
+        (*itr)->render();
     }
     glDepthFunc(GL_LESS);
 
-    renderfps();
+    if (!benchmode)
+        renderfps();
 
     glutSwapBuffers();
 }
 
-void idle() {
+namespace
+{
+    const std::string ftransformShader =
+        "#version 110\n"
+        "void main() {\n"
+        "    gl_Position = ftransform();\n"
+        "}\n";
 
-    static int ancient = 0;
-    static int last = 0;
-    static int msec = 0;        
-    last = msec;
-    msec = glutGet(GLUT_ELAPSED_TIME);
-    if (spin) {
-        rot += (msec-last)/10.0f; 
-        while (rot >= 360.0f)
-			rot -= 360.0f;
-    }
+    const std::string mvpCombinedShader =
+        "#version 110\n"
+        "void main() {\n"
+        "    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
+        "}\n";
 
-    static int fps = 0;
-    if (last / 1000 != msec / 1000) {
-        
-        float correctedFps = static_cast<float>(fps) * 1000.0f / static_cast<float>(msec - ancient);
-        fpsStream.str("");
-        fpsStream << "fps: " << correctedFps << std::ends;
-
-        ancient = msec;
-        fps = 0;
-    }
-
-    display();
-
-    ++fps;
-}
-
-void key(unsigned char k, int, int) {
-    switch (k) {
-    case ' ': 
-        spin = !spin; 
-        break;    
-    default:
-        break;
-    }
-    display();
+    const std::string mvpSeparateShader =
+        "#version 110\n"
+        "void main() {\n"
+        "    gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex;\n"
+        "}\n";
 }
 
 void menu(int value) {
@@ -377,34 +485,238 @@ void menu(int value) {
     case CSG_GRID2D:     setGrid2D();        break;
     case CSG_GRID3D:     setGrid3D();        break;
     case CSG_CUBERACK:   setCubeRack();      break;
+    case CSG_PIPE:       setPipe();          break;
     case CSG_CONCAVE:    setConcave();       break;
 
-    case ALGO_AUTOMATIC: OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::Automatic);
-                         break;
+    case ALGO_AUTOMATIC:
+        OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::Automatic);
+        break;
     case GF_STANDARD:    OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::Goldfeather);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::NoDepthComplexitySampling);
-                         break;
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::NoDepthComplexitySampling);
+        break;
     case GF_DC:          OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::Goldfeather);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::DepthComplexitySampling);
-                         break;
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::DepthComplexitySampling);
+        break;
     case GF_OQ:          OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::Goldfeather);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::OcclusionQuery);
-                         break;
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::OcclusionQuery);
+        break;
     case SCS_STANDARD:   OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::SCS);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::NoDepthComplexitySampling);
-                         break;
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::NoDepthComplexitySampling);
+        break;
     case SCS_DC:         OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::SCS);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::DepthComplexitySampling);
-                         break;
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::DepthComplexitySampling);
+        break;
     case SCS_OQ:         OpenCSG::setOption(OpenCSG::AlgorithmSetting, OpenCSG::SCS);
-                         OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::OcclusionQuery);
-                         break;
-
+        OpenCSG::setOption(OpenCSG::DepthComplexitySetting, OpenCSG::OcclusionQuery);
+        break;
+    case CAM_OUTSIDE_DEFAULT: inside = false;
+        OpenCSG::setOption(OpenCSG::CameraOutsideOptimization, OpenCSG::OptimizationDefault);
+        break;
+    case CAM_OUTSIDE_PERF: inside = false;
+        OpenCSG::setOption(OpenCSG::CameraOutsideOptimization, OpenCSG::OptimizationOn);
+        break;
+    case CAM_OUTSIDE_NOPERF: inside = false;
+        OpenCSG::setOption(OpenCSG::CameraOutsideOptimization, OpenCSG::OptimizationOff);
+        break;
+    case CAM_INSIDE:     inside = true;
+        OpenCSG::setOption(OpenCSG::CameraOutsideOptimization, OpenCSG::OptimizationOff);
+        break;
+    case CAM_INSIDE_DEFAULT: inside = true;
+        OpenCSG::setOption(OpenCSG::CameraOutsideOptimization, OpenCSG::OptimizationDefault);
+        break;
+    case GLSL_FIXED_FUNCTION:
+        OpenCSG::setVertexShader("");
+        break;
+    case GLSL_FTRANSFORM:
+        OpenCSG::setVertexShader(ftransformShader);
+        break;
+    case GLSL_MVP_COMBINED:
+        OpenCSG::setVertexShader(mvpCombinedShader);
+        break;
+    case GLSL_MVP_SEPARATE:
+        OpenCSG::setVertexShader(mvpSeparateShader);
+        break;
     default: break;
     }
     display();
 }
 
+void printNewBenchTableHeadline()
+{
+    const char* optionName = "";
+    switch (benchPerfOption)
+    {
+    case CAM_OUTSIDE_DEFAULT:
+        optionName = "Default performance";
+        break;
+    case CAM_OUTSIDE_PERF:
+        optionName = "Fast performance (at the expense of compatibility)";
+        break;
+    case CAM_OUTSIDE_NOPERF:
+        optionName = "Slower performance (but more compatible)";
+        break;
+    }
+    fprintf(stdout, "\n%s", optionName);
+    fprintf(stdout, "\n           Goldfeather    DC         OQ        SCS         DC         OQ");
+}
+
+void printNewBenchLine()
+{
+    const char* shapeName = "";
+    switch (benchShape)
+    {
+    case CSG_BASIC:
+        shapeName = "  Simple";
+        break;
+    case CSG_WIDGET:
+        shapeName = "  Widget";
+        break;
+    case CSG_GRID2D:
+        shapeName = " 2D-Grid";
+        break;
+    case CSG_GRID3D:
+        shapeName = " 3D-Grid";
+        break;
+    case CSG_CUBERACK:
+        shapeName = "Cuberack";
+        break;
+    case CSG_PIPE:
+        shapeName = "    Pipe";
+        break;
+    case CSG_CONCAVE:
+        shapeName = " Concave";
+        break;
+    }
+    fprintf(stdout, "\n%s:", shapeName);
+}
+
+void applyBenchSetting()
+{
+    menu(benchAlgorithm);
+    menu(benchShape);
+    menu(benchPerfOption);
+    benchSettingFrameOne = true;
+}
+
+void nextBenchSetting()
+{
+    if (benchShape == BENCH_START)
+    {
+        rot = 0.0f;
+        benchShape = CSG_BASIC;
+        benchAlgorithm = GF_STANDARD;
+        benchPerfOption = CAM_OUTSIDE_DEFAULT;
+        printNewBenchTableHeadline();
+        printNewBenchLine();
+        applyBenchSetting();
+        return;
+    }
+
+    ++benchAlgorithm;
+    if (   (benchShape != CSG_CONCAVE && benchAlgorithm <= SCS_OQ)
+        || (benchShape == CSG_CONCAVE && benchAlgorithm <= GF_OQ))
+    {
+        applyBenchSetting();
+        return;
+    }
+
+    benchAlgorithm = GF_STANDARD;
+    ++benchShape;
+    if (benchShape <= CSG_CONCAVE)
+    {
+        printNewBenchLine();
+        applyBenchSetting();
+        return;
+    }
+    fprintf(stdout, "\n");
+
+    benchShape = CSG_BASIC;
+    ++benchPerfOption;
+    if (benchPerfOption <= CAM_OUTSIDE_NOPERF)
+    {
+        printNewBenchTableHeadline();
+        printNewBenchLine();
+        applyBenchSetting();
+        return;
+    }
+
+    benchmode = false;
+}
+
+void idle() {
+
+    static int ancient = 0;
+    static int last = 0;
+    static int msec = 0;
+    static int numFramesRendered = 0;
+
+    last = msec;
+    msec = glutGet(GLUT_ELAPSED_TIME);
+    if (spin) {
+        rot += static_cast<float>(msec-last) / 10.0f;
+    }
+
+    if (benchSettingFrameOne)
+    {
+        rot = 0;
+        benchSettingFrameOne = false;
+        last = msec;
+        ancient = msec;
+        numFramesRendered = 0;
+    }
+
+    if (rot >= 360.0f)
+    {
+        while (rot >= 360.0f)
+            rot -= 360.0f;
+
+        if (benchmode)
+        {
+            float correctedFps = static_cast<float>(numFramesRendered) * 1000.0f / static_cast<float>(msec - ancient);
+            fprintf(stdout, "% 11.2f", correctedFps);
+            nextBenchSetting();
+
+            rot = 0.0f;
+            ancient = msec;
+            numFramesRendered = 0;
+        }
+    }
+
+    if (!benchmode)
+    {
+        if (last / 1000 != msec / 1000) {
+
+            float correctedFps = static_cast<float>(numFramesRendered) * 1000.0f / static_cast<float>(msec - ancient);
+            fpsStream.str("");
+            fpsStream << "fps: " << correctedFps << std::ends;
+
+            ancient = msec;
+            numFramesRendered = 0;
+        }
+    }
+
+    display();
+
+    ++numFramesRendered;
+}
+
+void key(unsigned char k, int, int) {
+    switch (k) {
+    case ' ':
+        spin = !spin;
+        break;
+    case 'b':
+        benchmode = true;
+        benchShape = BENCH_START;
+        nextBenchSetting();
+        fpsStream.str("");
+        break;
+    default:
+        break;
+    }
+    display();
+}
+
 void init()
 {
     // gray background
@@ -417,7 +729,7 @@ void init()
 
     glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
     glLightfv(GL_LIGHT0, GL_POSITION, light_position0);
-    glEnable(GL_LIGHT0);  
+    glEnable(GL_LIGHT0);
     glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
     glLightfv(GL_LIGHT1, GL_POSITION, light_position1);
     glEnable(GL_LIGHT1);
@@ -429,7 +741,7 @@ void init()
 
     // Setup the view of the CSG shape
     glMatrixMode(GL_PROJECTION);
-    gluPerspective(40.0, 1.0, 1.0, 10.0);
+    gluPerspective(40.0, 1.0, 0.2, 10.0);
     glMatrixMode(GL_MODELVIEW);
 
     // starting CSG shape
@@ -449,20 +761,36 @@ int main(int argc, char **argv)
     glutAddMenuEntry("2D-Grid",  CSG_GRID2D);
     glutAddMenuEntry("3D-Grid",  CSG_GRID3D);
     glutAddMenuEntry("Cuberack", CSG_CUBERACK);
+    glutAddMenuEntry("Pipe",     CSG_PIPE);
     glutAddMenuEntry("Concave",  CSG_CONCAVE);
-    
+
     int menuAlgorithm = glutCreateMenu(menu);
     glutAddMenuEntry("Automatic", ALGO_AUTOMATIC);
-    glutAddMenuEntry("Goldfeather standard",GF_STANDARD);
+    glutAddMenuEntry("Goldfeather standard", GF_STANDARD);
     glutAddMenuEntry("Goldfeather depth complexity sampling", GF_DC);
     glutAddMenuEntry("Goldfeather occlusion query", GF_OQ);
     glutAddMenuEntry("SCS standard", SCS_STANDARD);
     glutAddMenuEntry("SCS depth complexity sampling", SCS_DC);
     glutAddMenuEntry("SCS occlusion query", SCS_OQ);
 
+    int menuCamera = glutCreateMenu(menu);
+    glutAddMenuEntry("Camera outside (default performance optimizations)", CAM_OUTSIDE_DEFAULT);
+    glutAddMenuEntry("Camera outside (with performance optimizations)", CAM_OUTSIDE_PERF);
+    glutAddMenuEntry("Camera outside (without performance optimizations)", CAM_OUTSIDE_NOPERF);
+    glutAddMenuEntry("Camera inside (default performance optimizations)", CAM_INSIDE_DEFAULT);
+    glutAddMenuEntry("Camera inside (without performance optimizations)", CAM_INSIDE);
+
+    int menuGLSL = glutCreateMenu(menu);
+    glutAddMenuEntry("Fixed function", GLSL_FIXED_FUNCTION);
+    glutAddMenuEntry("ftransform()", GLSL_FTRANSFORM);
+    glutAddMenuEntry("gl_ModelViewProjectionMatrix * gl_Vertex", GLSL_MVP_COMBINED);
+    glutAddMenuEntry("gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex", GLSL_MVP_SEPARATE);
+
     glutCreateMenu(menu);
     glutAddSubMenu("CSG Shapes", menuShape);
     glutAddSubMenu("CSG Algorithms", menuAlgorithm);
+    glutAddSubMenu("Camera", menuCamera);
+    glutAddSubMenu("Vertex Shader", menuGLSL);
 
     // connect to right mouse button
     glutAttachMenu(GLUT_RIGHT_BUTTON);
@@ -476,4 +804,3 @@ int main(int argc, char **argv)
 
     return 0;
 }
-
