@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2010-2023, Florian Kirsch
+// Copyright (C) 2010-2025, Florian Kirsch
 //
 // This library is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
 // 51 Franklin Street, Fifth Floor, Boston, MA 02110 - 1301 USA.
 
 //
-// frameBufferObjectExt.h 
+// frameBufferObjectExt.h
 //
 // frame buffer object class implementing the offscreen buffer interface
 //
@@ -43,7 +43,7 @@ namespace OpenCSG {
 
             /// Initializes the frame buffer object with the intended width and height.
             /// The frame buffer object is created with RGBA and combined depth/stencil buffer.
-            virtual bool Initialize(int width, int height);
+            virtual bool Initialize(Dimensions dims);
 
             /// checks whether Initialize has been called before or not
             virtual bool IsInitialized() const { return initialized; }
@@ -51,7 +51,7 @@ namespace OpenCSG {
             /// Removes the frame buffer object OpenGL resources.
             virtual bool Reset();
             /// Change the size of the frame buffer object.
-            virtual bool Resize(int width, int height);
+            virtual bool Resize(Dimensions dims);
 
             /// Begin drawing to the frame buffer object. (i.e. use as "output" texture)
             virtual bool BeginCapture();
@@ -69,13 +69,13 @@ namespace OpenCSG {
             /// Returns the texture target this texture is bound to.
             virtual unsigned int GetTextureTarget() const { return textureTarget; }
             /// Returns the width of the frame buffer object.
-            virtual int GetWidth() const  { return width;  } 
+            virtual int GetWidth() const  { return dimensions.width;  }
             /// Returns the width of the frame buffer object.
-            virtual int GetHeight() const { return height; }
+            virtual int GetHeight() const { return dimensions.height; }
 
         protected:
-            int          width;     // width of the frame buffer object
-            int          height;    // height of the frame buffer object
+            // Width and height of the frame buffer object
+            Dimensions   dimensions;
 
             /// Texture stuff
             GLenum       textureTarget;
@@ -90,7 +90,6 @@ namespace OpenCSG {
         private:
             FrameBufferObjectExt(const FrameBufferObjectExt&);
             FrameBufferObjectExt& operator=(const FrameBufferObjectExt&);
-
         };
 
     } // namespace OpenGL
