@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2002-2023, Florian Kirsch,
+// Copyright (C) 2002-2025, Florian Kirsch,
 // Hasso-Plattner-Institute at the University of Potsdam, Germany
 //
 // This library is free software; you can redistribute it and/or modify
@@ -27,11 +27,14 @@
 #include "opencsgRender.h"
 #include "batch.h"
 #include "channelManager.h"
+#include "context.h"
 #include "occlusionQuery.h"
 #include "openglHelper.h"
 #include "primitiveHelper.h"
 #include "scissorMemo.h"
+#include "settings.h"
 #include <algorithm>
+#include <cassert>
 
 namespace OpenCSG {
 
@@ -46,8 +49,8 @@ namespace OpenCSG {
 
         void GoldfeatherChannelManager::merge() {
 
-            bool isFixedFunction = true;
-            setupProjectiveTexture(isFixedFunction);
+            ProjTextureSetup setup = FixedFunction;
+            setupProjectiveTexture(setup);
 
             glEnable(GL_ALPHA_TEST);
             glAlphaFunc(GL_GEQUAL, 0.5f); // accuracy issue with GL_EQUAL 1.0 on FX5600
@@ -74,7 +77,7 @@ namespace OpenCSG {
                         (*j)->render();
                     }
                 } else {
-                    // shapes of interest: we need to determine the appropriate layer of 
+                    // shapes of interest: we need to determine the appropriate layer of
                     // the shapes, using stencil counting
                     glClearStencil(0);
                     glStencilMask(OpenGL::stencilMask);
@@ -90,12 +93,133 @@ namespace OpenCSG {
 
             scissor->disableScissor();
 
-            resetProjectiveTexture(isFixedFunction);
+            resetProjectiveTexture(setup);
 
             clear();
         }
 
-        GoldfeatherChannelManager* channelMgr;
+
+        class GoldfeatherChannelManagerGLSLProgram : public ChannelManagerForBatches {
+        public:
+            virtual void merge();
+        };
+
+
+        static const char mergeFragmentProgramRect[] =
+            "#version 110\n"
+            "#extension GL_ARB_texture_rectangle : enable\n"
+            "uniform sampler2DRect texture0;\n"
+            "uniform vec4 color;\n"
+            "void main() {\n"
+            "    vec4 temp = texture2DRect(texture0, gl_FragCoord.xy);\n"
+            "    float d = dot(temp, color);\n"
+            "    if (d < 0.5)\n"
+            "        discard;\n"
+            "    gl_FragColor = color;\n"
+            "}\n";
+
+        static const char mergeFragmentProgram2D[] =
+            "#version 110\n"
+            "uniform sampler2D texture0;\n"
+            "uniform vec2 texSizeInv;\n"
+            "uniform vec4 color;\n"
+            "void main() {\n"
+            "    vec2 texCoord = vec2(gl_FragCoord.x * texSizeInv.x, gl_FragCoord.y * texSizeInv.y);\n"
+            "    vec4 temp = texture2D(texture0, texCoord);\n"
+            "    float d = dot(temp, color);\n"
+            "    if (d < 0.5)\n"
+            "        discard;\n"
+            "    gl_FragColor = color;\n"
+            "}\n";
+
+        void GoldfeatherChannelManagerGLSLProgram::merge()
+        {
+            const int GFIdOffset = 0;
+            const char* programID = getVertexShader() + (isRectangularTexture() ? 1 : 0) + GFIdOffset;
+            GLuint glslProgram =
+                isRectangularTexture()
+                  ? OpenGL::getGLSLProgram(programID, getVertexShader(), mergeFragmentProgramRect)
+                  : OpenGL::getGLSLProgram(programID, getVertexShader(), mergeFragmentProgram2D);
+
+            GLint col = glGetUniformLocation(glslProgram, "color");
+            GLint texSizeInv = -1;
+            if (!isRectangularTexture())
+                texSizeInv = glGetUniformLocation(glslProgram, "texSizeInv");
+
+            glUseProgram(glslProgram);
+
+            ProjTextureSetup setup = GLSLProgram;
+            setupProjectiveTexture(setup, texSizeInv);
+
+            glDisable(GL_ALPHA_TEST);
+            glEnable(GL_CULL_FACE);
+            glEnable(GL_DEPTH_TEST);
+            glDepthFunc(GL_LESS);
+            glDepthMask(GL_TRUE);
+            glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
+
+            std::vector<Channel> channels = occupied();
+            for (std::vector<Channel>::const_iterator c = channels.begin(); c != channels.end(); ++c) {
+
+                const std::vector<Primitive*> primitives = getPrimitives(*c);
+
+                scissor->recall(*c);
+                scissor->enableScissor();
+
+                GLfloat refColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
+
+                switch (*c) {
+                case Alpha:
+                    refColor[3] = 1.0f;
+                    break;
+                case Red:
+                    refColor[0] = 1.0f;
+                    break;
+                case Green:
+                    refColor[1] = 1.0f;
+                    break;
+                case Blue:
+                    refColor[2] = 1.0f;
+                    break;
+                default:
+                    // should not happen!
+                    assert(0);
+                };
+
+                glUniform4fv(col, 1, refColor);
+
+                if (getLayer(*c) == -1) {
+
+                    glEnable(GL_CULL_FACE);
+                    for (Batch::const_iterator j = primitives.begin(); j != primitives.end(); ++j) {
+                        glCullFace((*j)->getOperation() == Intersection ? GL_BACK : GL_FRONT);
+                        (*j)->render();
+                    }
+                }
+                else {
+                    // shapes of interest: we need to determine the appropriate layer of
+                    // the shapes, using stencil counting
+                    glClearStencil(0);
+                    glStencilMask(OpenGL::stencilMask);
+                    glClear(GL_STENCIL_BUFFER_BIT);
+                    OpenGL::renderLayer(getLayer(*c), primitives);
+                    glDisable(GL_STENCIL_TEST);
+                }
+            }
+
+            scissor->disableScissor();
+
+            glDisable(GL_CULL_FACE);
+            glDepthFunc(GL_LEQUAL);
+            glUseProgram(0);
+
+            resetProjectiveTexture(setup);
+
+            clear();
+        }
+
+
+        ChannelManagerForBatches* channelMgr;
 
         void touchFragments(const Batch& batch) {
             glMatrixMode(GL_PROJECTION);
@@ -136,6 +260,15 @@ namespace OpenCSG {
             glPopMatrix();
         }
 
+        GLenum getParityDepthFunc()
+        {
+            int setting = getOption(CameraOutsideOptimization);
+            if (setting == OptimizationOn || setting == OptimizationForceOn)
+                return GL_LEQUAL;
+            else
+                return GL_GREATER;
+        }
+
         void discardFragments(const Batch& batch, int parity, int mask) {
             glDepthFunc(GL_ALWAYS);
             glDepthMask(GL_TRUE);
@@ -150,7 +283,7 @@ namespace OpenCSG {
 
             glDepthRange(0.0, 1.0);
             glDepthMask(GL_FALSE);
-            glDepthFunc(GL_LEQUAL);
+            glDepthFunc(getParityDepthFunc());
         }
 
         void discardFragments(int parity, int mask) {
@@ -167,7 +300,23 @@ namespace OpenCSG {
 
             glDepthRange(0.0, 1.0);
             glDepthMask(GL_FALSE);
-            glDepthFunc(GL_LEQUAL);
+            glDepthFunc(getParityDepthFunc());
+        }
+
+        void discardFragments(
+                const Batch& batch,
+                int parity, int mask,
+                bool layered) {
+
+            if (layered) {
+                discardFragments(parity, mask);
+            } else {
+                if (scissor->isDepthBoundsTestEnabled())
+                    glDisable(GL_DEPTH_BOUNDS_TEST_EXT);
+                discardFragments(batch, parity, mask);
+                if (scissor->isDepthBoundsTestEnabled())
+                    glEnable(GL_DEPTH_BOUNDS_TEST_EXT);
+            }
         }
 
         void parityTestAndDiscard(
@@ -177,7 +326,7 @@ namespace OpenCSG {
                 unsigned int stencilMax) {
 
             glDepthMask(GL_FALSE);
-            glDepthFunc(GL_LEQUAL);
+            glDepthFunc(getParityDepthFunc());
 
             glEnable(GL_STENCIL_TEST);
 
@@ -187,7 +336,10 @@ namespace OpenCSG {
             // for all shapes of the intersection, we conduct the parity test.
             // for fragments for which it fails, we mark them as not visible
             // (parity testing means to check whether the number of surfaces in front
-            // of the current z-buffer is even or uneven)
+            // of the current z-buffer is even or uneven. As all shapes are closed,
+            // this is equivalent to check the number of surfaces behind the current
+            // z-buffer. This is what we effectively do by default, because
+            // that approach is more robust)
             for (std::vector<Primitive*>::const_iterator itr = primitives.begin(); itr != primitives.end(); ++itr ) {
                 if (!layered) {
 
@@ -207,7 +359,7 @@ namespace OpenCSG {
                     }
 
                     // for substracted shapes that don't touch the shapes of interest,
-                    // the parity test would always fail. thus, they are omited here. 
+                    // the parity test would always fail. thus, they are omited here.
                     bool needParityTest = ((*itr)->getOperation() == Intersection);
                     if (!needParityTest) {
                         for (Batch::const_iterator k = shapesOfInterest.begin(); k != shapesOfInterest.end(); ++k) {
@@ -217,8 +369,8 @@ namespace OpenCSG {
                             }
                         }
                     }
-                    if (!needParityTest) continue;
-
+                    if (!needParityTest)
+                        continue;
                 }
 
                 // we only need one bit in the stencil buffer for each parity test.
@@ -226,16 +378,13 @@ namespace OpenCSG {
                 // used, we discard fragments marked invisible by the parity test,
                 // and clear the stencil buffer.
                 if (parityValue >= stencilMax) {
-                    if (layered) {
-                        discardFragments(allParityTestValues, parityValue - 1);
-                    } else {
-                        discardFragments(shapesOfInterest, allParityTestValues, parityValue - 1);
-                    }
+                    discardFragments(shapesOfInterest, allParityTestValues, parityValue - 1, layered);
                     parityValue = 1;
                     allParityTestValues = 0;
                 }
 
-                // parity test: count surfaces in front of shapes of interest
+                // parity test: count surfaces behind (or, depending on setting, in front)
+                // of shapes of interest, i.e., the current z-buffer.
                 channelMgr->renderToChannel(false);
                 glStencilFunc(GL_ALWAYS, 0, parityValue);
                 glStencilMask(parityValue);
@@ -250,21 +399,16 @@ namespace OpenCSG {
                 //   if #(surfaces) is uneven (stencilValue == parityValue):
                 //      if intersecting shapes      -> shapes intersect, first shape is visible, don't change depth
                 //      if subtracting second shape -> front surface of first shape is subtracted, not visible. set depth to max
-                //if (!(*i)->complement_) {
                 if ((*itr)->getOperation() == Intersection) {
                     allParityTestValues += parityValue;
                 }
-        
+
                 parityValue <<= 1;
-            }     
-    
+            }
+
             // discard fragments marked invisible the last parity tests (and clear the stencil buffer)
             if (parityValue != 1) { // that would mean no parity test had occured at all
-                if (layered) {
-                    discardFragments(allParityTestValues, parityValue - 1);
-                } else {
-                    discardFragments(shapesOfInterest, allParityTestValues, parityValue - 1);
-                }
+                discardFragments(shapesOfInterest, allParityTestValues, parityValue - 1, layered);
             }
 
             glDisable(GL_STENCIL_TEST);
@@ -321,7 +465,11 @@ namespace OpenCSG {
                     glClear(GL_STENCIL_BUFFER_BIT);
                 }
 
-                scissor->enableDepthBoundsBack();
+                GLenum depthFunc = getParityDepthFunc();
+                if (depthFunc == GL_GREATER)
+                    scissor->enableDepthBoundsFront();
+                else if (depthFunc == GL_LEQUAL)
+                    scissor->enableDepthBoundsBack();
 
                 parityTestAndDiscard(*itr, primitives, false, OpenGL::stencilMask);
 
@@ -460,9 +608,22 @@ namespace OpenCSG {
         delete scissor;
     }
 
+    ChannelManagerForBatches* getChannelManager() {
+
+        if (GLAD_GL_VERSION_2_0)
+        {
+            bool useGLSL = getVertexShader() != 0;
+            if (useGLSL)
+                return new GoldfeatherChannelManagerGLSLProgram;
+        }
+
+        return new GoldfeatherChannelManager;
+    }
+
     void renderGoldfeather(const std::vector<Primitive*>& primitives, DepthComplexityAlgorithm algorithm)
     {
-        channelMgr = new GoldfeatherChannelManager;
+        channelMgr = getChannelManager();
+
         if (channelMgr->init())
         {
             switch (algorithm) {
@@ -478,8 +639,6 @@ namespace OpenCSG {
             case DepthComplexitySampling:
                 renderDepthComplexitySamplingGoldfeather(primitives);
                 break;
-            case DepthComplexityAlgorithmUnused:
-                break; // does not happen
             }
         }
 
