@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2002-2023, Florian Kirsch,
+// Copyright (C) 2002-2025, Florian Kirsch,
 // Hasso-Plattner-Institute at the University of Potsdam, Germany
 //
 // This library is free software; you can redistribute it and/or modify
@@ -32,6 +32,8 @@
 #include "openglHelper.h"
 #include "primitiveHelper.h"
 #include "scissorMemo.h"
+#include "sequencer.h"
+#include "settings.h"
 
 #include <algorithm>
 #include <map>
@@ -71,8 +73,8 @@ namespace OpenCSG {
 
         void SCSChannelManagerAlphaOnly::merge() {
 
-            bool isFixedFunction = true;
-            setupProjectiveTexture(isFixedFunction);
+            ProjTextureSetup setup = FixedFunction;
+            setupProjectiveTexture(setup);
 
             glEnable(GL_ALPHA_TEST);
             glEnable(GL_CULL_FACE);
@@ -130,20 +132,21 @@ namespace OpenCSG {
             glDisable(GL_CULL_FACE);
             glDepthFunc(GL_LEQUAL);
 
-            resetProjectiveTexture(isFixedFunction);
+            resetProjectiveTexture(setup);
 
             clear();
         }
 
         // Stores Ids in the all components of the color buffer
         // -> in theory, 2^32-1 primitives are possible
-        class SCSChannelManagerFragmentProgram : public ChannelManagerForBatches {
+        // ARB variant
+        class SCSChannelManagerARBProgram : public ChannelManagerForBatches {
         public:
             virtual Channel request();
             virtual void merge();
         };
 
-        Channel SCSChannelManagerFragmentProgram::request() {
+        Channel SCSChannelManagerARBProgram::request() {
             ChannelManagerForBatches::request();
             mCurrentChannel = AllChannels;
             mOccupiedChannels = mCurrentChannel;
@@ -151,7 +154,7 @@ namespace OpenCSG {
         }
 
         // Emulates eye texgen
-        static const char mergeVertexProgram[] =
+        static const char mergeARBVertexProgram[] =
 "!!ARBvp1.0 OPTION ARB_position_invariant;\n"
 "ATTRIB  pos = vertex.position;\n"
 "ATTRIB  col = vertex.color;\n"
@@ -186,7 +189,7 @@ namespace OpenCSG {
         // in contract to the GL_EQUAL in SCSChannelManagerAlphaOnly above.
         // The scaling by 2.0f is required for NVidia hardware, which considers
         // the alpha function GL_LESS, 0.5f/255.0f as GL_LESS, 0.0f for some reason.
-        static const char mergeFragmentProgramRect[] =
+        static const char mergeARBFragmentProgramRect[] =
 "!!ARBfp1.0\n"
 "TEMP    temp;\n"
 "ATTRIB  tex0 = fragment.texcoord[0];\n"
@@ -199,7 +202,7 @@ namespace OpenCSG {
 "DP4     out, temp, scaleByTwo;\n"
 "END";
 
-        static const char mergeFragmentProgram2D[] =
+        static const char mergeARBFragmentProgram2D[] =
 "!!ARBfp1.0\n"
 "TEMP    temp;\n"
 "ATTRIB  tex0 = fragment.texcoord[0];\n"
@@ -212,21 +215,21 @@ namespace OpenCSG {
 "DP4     out, temp, scaleByTwo;\n"
 "END";
 
-        void SCSChannelManagerFragmentProgram::merge()
+        void SCSChannelManagerARBProgram::merge()
         {
-            GLuint vId = OpenGL::getARBVertexProgram(mergeVertexProgram, (sizeof(mergeVertexProgram) / sizeof(mergeVertexProgram[0])) - 1);
+            GLuint vId = OpenGL::getARBVertexProgram(mergeARBVertexProgram, (sizeof(mergeARBVertexProgram) / sizeof(mergeARBVertexProgram[0])) - 1);
             glBindProgramARB(GL_VERTEX_PROGRAM_ARB, vId);
             glEnable(GL_VERTEX_PROGRAM_ARB);
 
             GLuint fId =
-                 isRectangularTexture()
-                   ? OpenGL::getARBFragmentProgram(mergeFragmentProgramRect, (sizeof(mergeFragmentProgramRect) / sizeof(mergeFragmentProgramRect[0])) - 1)
-                   : OpenGL::getARBFragmentProgram(mergeFragmentProgram2D, (sizeof(mergeFragmentProgram2D) / sizeof(mergeFragmentProgram2D[0])) - 1);
+                isRectangularTexture()
+                  ? OpenGL::getARBFragmentProgram(mergeARBFragmentProgramRect, (sizeof(mergeARBFragmentProgramRect) / sizeof(mergeARBFragmentProgramRect[0])) - 1)
+                  : OpenGL::getARBFragmentProgram(mergeARBFragmentProgram2D, (sizeof(mergeARBFragmentProgram2D) / sizeof(mergeARBFragmentProgram2D[0])) - 1);
             glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fId);
             glEnable(GL_FRAGMENT_PROGRAM_ARB);
 
-            bool isFixedFunction = false;
-            setupProjectiveTexture(isFixedFunction);
+            ProjTextureSetup setup = ARBShader;
+            setupProjectiveTexture(setup);
 
             glEnable(GL_ALPHA_TEST);
             glEnable(GL_CULL_FACE);
@@ -262,7 +265,113 @@ namespace OpenCSG {
             glDisable(GL_FRAGMENT_PROGRAM_ARB);
             glDisable(GL_VERTEX_PROGRAM_ARB);
 
-            resetProjectiveTexture(isFixedFunction);
+            resetProjectiveTexture(setup);
+
+            clear();
+        }
+
+        // Stores Ids in the all components of the color buffer
+        // -> in theory, 2^32-1 primitives are possible
+        // GLSL variant
+        class SCSChannelManagerGLSLProgram : public ChannelManagerForBatches {
+        public:
+            virtual Channel request();
+            virtual void merge();
+        };
+
+        Channel SCSChannelManagerGLSLProgram::request() {
+            ChannelManagerForBatches::request();
+            mCurrentChannel = AllChannels;
+            mOccupiedChannels = mCurrentChannel;
+            return mCurrentChannel;
+        }
+
+        // Subtract color from texture value, and if the result is not 0
+        // (including some epsilon, which is less than 1.0/(256*256),
+        // discards the fragment.
+        // This way, all 32-bits of the color channel can be used
+        // for an 'equal' alpha test, i.e, to check if value in texture
+        // and color are equal.
+        static const char mergeFragmentProgramRect[] =
+            "#version 110\n"
+            "#extension GL_ARB_texture_rectangle : enable\n"
+            "uniform sampler2DRect texture0;\n"
+            "uniform vec4 color;\n"
+            "void main() {\n"
+            "    vec4 temp = texture2DRect(texture0, gl_FragCoord.xy);\n"
+            "    temp = temp - color;\n"
+            "    if (dot(temp, temp) > 0.000001)\n"
+            "        discard;\n"
+            "    gl_FragColor = color;\n"
+            "}\n";
+
+        static const char mergeFragmentProgram2D[] =
+            "#version 110\n"
+            "uniform sampler2D texture0;\n"
+            "uniform vec2 texSizeInv;\n"
+            "uniform vec4 color;\n"
+            "void main() {\n"
+            "    vec2 texCoord = vec2(gl_FragCoord.x * texSizeInv.x, gl_FragCoord.y * texSizeInv.y);\n"
+            "    vec4 temp = texture2D(texture0, texCoord);\n"
+            "    temp = temp - color;\n"
+            "    if (dot(temp, temp) > 0.000001)\n"
+            "        discard;\n"
+            "    gl_FragColor = color;\n"
+            "}\n";
+
+        void SCSChannelManagerGLSLProgram::merge()
+        {
+            const int SCSIdOffset = 2;
+            const char* programID = getVertexShader() + (isRectangularTexture() ? 1 : 0) + SCSIdOffset;
+            GLuint glslProgram =
+                isRectangularTexture()
+                    ? OpenGL::getGLSLProgram(programID, getVertexShader(), mergeFragmentProgramRect)
+                    : OpenGL::getGLSLProgram(programID, getVertexShader(), mergeFragmentProgram2D);
+
+            GLint col = glGetUniformLocation(glslProgram, "color");
+            GLint texSizeInv = -1;
+            if (!isRectangularTexture())
+                texSizeInv = glGetUniformLocation(glslProgram, "texSizeInv");
+
+            glUseProgram(glslProgram);
+
+            ProjTextureSetup setup = GLSLProgram;
+            setupProjectiveTexture(setup, texSizeInv);
+
+            glDisable(GL_ALPHA_TEST);
+            glEnable(GL_CULL_FACE);
+            glEnable(GL_DEPTH_TEST);
+            glDepthFunc(GL_LESS);
+            glDepthMask(GL_TRUE);
+            glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
+
+            std::vector<Channel> channels;
+            channels.push_back(AllChannels);
+            for (std::vector<Channel>::const_iterator c = channels.begin(); c!=channels.end(); ++c) {
+
+                scissor->recall(*c);
+                scissor->enableScissor();
+
+                const std::vector<Primitive*> primitives = getPrimitives(*c);
+                for (std::vector<Primitive*>::const_iterator j = primitives.begin(); j != primitives.end(); ++j) {
+                    glCullFace((*j)->getOperation() == Intersection ? GL_BACK : GL_FRONT);
+                    RenderData* primitiveData = getRenderData(*j);
+                    GLubyte * id = primitiveData->bufferId.vec();
+                    glUniform4f(col, static_cast<float>(static_cast<double>(id[0]) / 255.0),
+                                     static_cast<float>(static_cast<double>(id[1]) / 255.0),
+                                     static_cast<float>(static_cast<double>(id[2]) / 255.0),
+                                     static_cast<float>(static_cast<double>(id[3]) / 255.0));
+                    (*j)->render();
+                }
+            }
+
+            scissor->disableScissor();
+
+            glDisable(GL_CULL_FACE);
+            glDepthFunc(GL_LEQUAL);
+            glUseProgram(0);
+
+            resetProjectiveTexture(setup);
 
             clear();
         }
@@ -270,12 +379,27 @@ namespace OpenCSG {
 
         ChannelManagerForBatches* getChannelManager() {
 
+            if (GLAD_GL_VERSION_2_0)
+            {
+                bool useGLSL = getVertexShader() != 0;
+                if (useGLSL)
+                    return new SCSChannelManagerGLSLProgram;
+            }
+
+            // The ARB vertex program path has the following problem:
+            // With Nouveau drivers on Linux and an NVidia GTX 710, the ARB_position_invariant
+            // option appears to be buggy sometimes, and this causes z-buffer artifacts.
             if (   OPENCSG_HAS_EXT(ARB_vertex_program)
                 && OPENCSG_HAS_EXT(ARB_fragment_program)
             ) {
-                return new SCSChannelManagerFragmentProgram;
+                return new SCSChannelManagerARBProgram;
             }
 
+            // The fallback path, using fixed function OpenGL, has its own problems:
+            // Obviously, it does not support more than 255 primitives. More importantly,
+            // due to the alpha test and the GL_EQUAL alpha function used here, due
+            // to accuracy problems, the alpha test for some IDs may incorrectly fail.
+            // Then, some primitives of the CSG share are completely missing.
             return new SCSChannelManagerAlphaOnly;
         }
 
@@ -311,7 +435,7 @@ namespace OpenCSG {
                 glDepthFunc(GL_GREATER);
                 glCullFace(GL_BACK);
                 glEnable(GL_CULL_FACE);
-                RenderData* primitiveData = getRenderData(primitives[0]);
+                RenderData * primitiveData = getRenderData(primitives[0]);
                 GLubyte * id = primitiveData->bufferId.vec();
                 glColor4ubv(id);
                 primitives[0]->render();
@@ -330,7 +454,7 @@ namespace OpenCSG {
 
             {
                 for (std::vector<Primitive*>::const_iterator i = primitives.begin(); i != primitives.end(); ++i) {
-                    RenderData* primitiveData = getRenderData(*i);
+                    RenderData * primitiveData = getRenderData(*i);
                     GLubyte * id = primitiveData->bufferId.vec();
                     glColor4ubv(id);
                     (*i)->render();
@@ -369,23 +493,37 @@ namespace OpenCSG {
             glDisable(GL_STENCIL_TEST);
         }
 
-        void subtractPrimitives(std::vector<Batch>::const_iterator begin, 
-                                std::vector<Batch>::const_iterator end,
-                                const unsigned int iterations) {
+        void subtractPrimitives(const std::vector<Batch>& batches,
+                                const unsigned int depthComplexity = 0) {
 
-            if (begin == end) {
-                return;
-            }
+            int setting = getOption(CameraOutsideOptimization);
+            bool cameraInsideModel = (setting == OptimizationOff);
 
             glStencilMask(OpenGL::stencilMask);
             glEnable(GL_STENCIL_TEST);
             glEnable(GL_CULL_FACE);
 
+            size_t numberOfBatches = batches.size();
+            BouncingSequencer bounce(numberOfBatches);
+            SchoenfieldSequencer schoenfield(numberOfBatches);
+            Sequencer * sequencer = 0;
+            size_t numIterations;
+            if (depthComplexity == 0)
+            {
+                sequencer = &schoenfield;
+                numIterations = sequencer->size();
+            }
+            else
+            {
+                sequencer = &bounce;
+                numIterations = sequencer->sizeForDepthComplexity(depthComplexity);
+            }
+
             unsigned int stencilref = 0;
-            int sense = 1;
-            unsigned int changes = 0;
-            std::vector<Batch>::const_iterator i = begin;
-            do {
+            for (size_t i = 0; i < numIterations; ++i)
+            {
+                const Batch& batch = batches[sequencer->index(i)];
+
                 // create a distinct reference value
                 ++stencilref;
                 if (stencilref == OpenGL::stencilMax) {
@@ -394,14 +532,29 @@ namespace OpenCSG {
                 }
 
                 channelMgr->renderToChannel(false);
-                glDepthFunc(GL_LESS);
-                glDepthMask(GL_FALSE);
+
+                glDepthMask(GL_FALSE );
                 glStencilFunc(GL_ALWAYS, stencilref, OpenGL::stencilMask);
                 glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
+
+                if (cameraInsideModel)
+                {
+                    glDepthFunc(GL_GREATER);
+                    glCullFace(GL_FRONT);
+
+                    for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
+                        (*j)->render();
+                    }
+
+                    glStencilFunc(GL_EQUAL, stencilref, OpenGL::stencilMask);
+                    glStencilOp(GL_ZERO, GL_ZERO, GL_KEEP);
+                }
+
+                glDepthFunc(GL_LESS);
                 glCullFace(GL_BACK);
 
-                { 
-                    for (Batch::const_iterator j = i->begin(); j != i->end(); ++j) {
+                {
+                    for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
                         (*j)->render();
                     }
                 }
@@ -415,62 +568,47 @@ namespace OpenCSG {
                 glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
 
                 {
-                    for (Batch::const_iterator j = i->begin(); j != i->end(); ++j) {
-                        RenderData* primitiveData = getRenderData(*j);
+                    for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
+                        RenderData * primitiveData = getRenderData(*j);
                         GLubyte * id = primitiveData->bufferId.vec();
                         glColor4ubv(id);
                         (*j)->render();
                     }
                 }
-
-                if (sense == 1) {
-                    ++i; 
-                    if (i == end) {
-                        sense = 0;
-                        --i; if (i == begin) break; // only one subtracted shape
-                        --i;
-                        ++changes;
-                    }
-                } else {
-                    if (i == begin) {
-                        sense = 1;
-                        ++i;
-                        ++changes;
-                    } else {
-                        --i;
-                    }
-                }
-            } while (changes < iterations);
+            };
 
             glDisable(GL_STENCIL_TEST);
         }
 
-        bool subtractPrimitivesWithOcclusionQueries(std::vector<Batch>::const_iterator begin,
-                                                    std::vector<Batch>::const_iterator end) {
-
-            const std::size_t numberOfBatches = end - begin;
-            if (numberOfBatches == 0) {
-                return true;
-            }
+        bool subtractPrimitivesWithOcclusionQueries(const std::vector<Batch>& batches) {
 
             OpenGL::OcclusionQuery* occlusionTest = OpenGL::getOcclusionQuery(true);
             if (!occlusionTest) {
                 return false;
             }
 
+            const std::size_t numberOfBatches = batches.size();
+
+            int setting = getOption(CameraOutsideOptimization);
+            bool cameraInsideModel = (setting == OptimizationOff);
+
             glStencilMask(OpenGL::stencilMask);
             glEnable(GL_STENCIL_TEST);
             glEnable(GL_CULL_FACE);
 
             std::vector<unsigned int> fragmentcount(numberOfBatches, 0);
+            unsigned int shapesWithoutUpdate = 0;
+
+            SimpleSequencer sequencer(numberOfBatches);
+            size_t numIterations = sequencer.size();
 
             unsigned int stencilref = 0;
-            std::vector<Batch>::const_iterator i = begin;
-            unsigned int shapesWithoutUpdate = 0;    
-            unsigned int shapeCount = 0;
-            int idx=0;
 
-            do {
+            for (size_t i = 0; i < numIterations; ++i)
+            {
+                size_t idx = sequencer.index(i);
+                const Batch& batch = batches[idx];
+
                 // create a distinct reference value
                 ++stencilref;
                 if (stencilref == OpenGL::stencilMax) {
@@ -479,15 +617,32 @@ namespace OpenCSG {
                 }
 
                 channelMgr->renderToChannel(false);
-                glDepthFunc(GL_LESS);
+
                 glDepthMask(GL_FALSE);
                 glStencilFunc(GL_ALWAYS, stencilref, OpenGL::stencilMask);
                 glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
-                glCullFace(GL_BACK);
 
                 occlusionTest->beginQuery();
-                { 
-                    for (Batch::const_iterator j = i->begin(); j != i->end(); ++j) {
+                if (cameraInsideModel)
+                {
+                    glDepthFunc(GL_GREATER);
+                    glCullFace(GL_FRONT);
+
+                    {
+                        for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
+                            (*j)->render();
+                        }
+                    }
+
+                    glStencilFunc(GL_EQUAL, stencilref, OpenGL::stencilMask);
+                    glStencilOp(GL_ZERO, GL_ZERO, GL_KEEP);
+                }
+
+                glDepthFunc(GL_LESS);
+                glCullFace(GL_BACK);
+
+                {
+                    for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
                         (*j)->render();
                     }
                 }
@@ -503,8 +658,8 @@ namespace OpenCSG {
                 glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
 
                 {
-                    for (Batch::const_iterator j = i->begin(); j != i->end(); ++j) {
-                        RenderData* primitiveData = getRenderData(*j);
+                    for (Batch::const_iterator j = batch.begin(); j != batch.end(); ++j) {
+                        RenderData * primitiveData = getRenderData(*j);
                         GLubyte * id = primitiveData->bufferId.vec();
                         glColor4ubv(id);
                         (*j)->render();
@@ -517,17 +672,10 @@ namespace OpenCSG {
                     shapesWithoutUpdate = 0;
                 } else {
                     ++shapesWithoutUpdate;
+                    if (shapesWithoutUpdate >= numberOfBatches)
+                        break;
                 }
-
-                ++i; ++idx; if (i == end) {
-                    i = begin;
-                    idx = 0;
-                }
-    
-                ++shapeCount;
-                if (shapeCount >= (numberOfBatches) * (numberOfBatches) - (numberOfBatches) + 1) break;
-
-            } while (shapesWithoutUpdate < numberOfBatches);
+            }
 
             delete occlusionTest;
 
@@ -539,7 +687,7 @@ namespace OpenCSG {
         void renderIntersectedBack(const std::vector<Primitive*>& primitives) {
             // where a back face of intersected shape is in front of any subtracted shape
             // mask fragment as invisible. Updating depth values is not necessary, so when
-            // having IDs, this is kind of simple. 
+            // having IDs, this is kind of simple.
             channelMgr->renderToChannel(true);
             glEnable(GL_CULL_FACE);
             glCullFace(GL_FRONT);
@@ -598,8 +746,8 @@ namespace OpenCSG {
         if (algorithm == DepthComplexitySampling) {
             scissor->enableScissor();
             glClear(GL_STENCIL_BUFFER_BIT);
-            depthComplexity = 
-                (std::min)(OpenGL::calcMaxDepthComplexity(subtracted, scissor->getCurrentArea()), 
+            depthComplexity =
+                (std::min)(OpenGL::calcMaxDepthComplexity(subtracted, scissor->getCurrentArea()),
                            static_cast<unsigned int>(subtractedBatches.size()));
         }
 
@@ -617,23 +765,24 @@ namespace OpenCSG {
         glClearDepth(1.0);
 
         renderIntersectedFront(intersected);
-        scissor->enableDepthBounds();
-        switch (algorithm) {
-        case OcclusionQuery:
-            if (subtractPrimitivesWithOcclusionQueries(subtractedBatches.begin(), subtractedBatches.end()))
-                break; // success
-            // Maybe we just should give up here?
-            // fall through
-        case NoDepthComplexitySampling:
-            subtractPrimitives(subtractedBatches.begin(), subtractedBatches.end(), static_cast<unsigned int>(subtractedBatches.size()));
-            break;
-        case DepthComplexitySampling:
-            subtractPrimitives(subtractedBatches.begin(), subtractedBatches.end(), depthComplexity);
-            break;
-        case DepthComplexityAlgorithmUnused:
-            break; // does not happen when invoked correctly           
+        if (!subtractedBatches.batches().empty())
+        {
+            scissor->enableDepthBounds();
+            switch (algorithm) {
+            case OcclusionQuery:
+                if (subtractPrimitivesWithOcclusionQueries(subtractedBatches.batches()))
+                    break; // success
+                // Maybe we just should give up here?
+                // fall through
+            case NoDepthComplexitySampling:
+                subtractPrimitives(subtractedBatches.batches());
+                break;
+            case DepthComplexitySampling:
+                subtractPrimitives(subtractedBatches.batches(), depthComplexity);
+                break;
+            }
+            scissor->disableDepthBounds();
         }
-        scissor->disableDepthBounds();
         renderIntersectedBack(intersected);
 
         scissor->disableScissor();
