@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2002-2023, Florian Kirsch,
+// Copyright (C) 2002-2025, Florian Kirsch,
 // Hasso-Plattner-Institute at the University of Potsdam, Germany
 //
 // This library is free software; you can redistribute it and/or modify
@@ -17,7 +17,7 @@
 // 51 Franklin Street, Fifth Floor, Boston, MA 02110 - 1301 USA.
 
 //
-// scissorMemo.h 
+// scissorMemo.h
 //
 // handles scissor calculations
 //
@@ -37,7 +37,7 @@ namespace OpenCSG {
         /// convenience class for handling scissoring. All areas
         /// are given in normal device coordinates [-1, 1] x [-1, 1].
         ScissorMemo();
-        
+
         /// for each channel, the currently set area can be stored
         void store(Channel);
         /// restores the area for the channel
@@ -48,10 +48,14 @@ namespace OpenCSG {
         /// disables scissoring
         void disableScissor() const;
 
+        /// checks if the depth bounds test is enabled
+        bool isDepthBoundsTestEnabled() const;
         /// enables depth bounds test of volume (if supported by graphics hardware)
         void enableDepthBounds() const;
         /// enables depth bounds test of back of current volume (if supported by graphics hardware)
         void enableDepthBoundsBack() const;
+        /// enables depth bounds test of front of current volume (if supported by graphics hardware)
+        void enableDepthBoundsFront() const;
         /// disables depth bounds test of volume (if supported by graphics hardware)
         void disableDepthBounds() const;
 
@@ -73,7 +77,7 @@ namespace OpenCSG {
         void calculateArea();
 
     private:
-        NDCVolume mIintersection;
+        NDCVolume mIntersection;
         NDCVolume mCurrent;
         NDCVolume mArea;
         std::vector<NDCVolume> mScissor;
@@ -84,5 +88,3 @@ namespace OpenCSG {
 } // namespace OpenCSG
 
 #endif // __OpenCSG__scissor_memo_h__
-
-
