@@ -1,5 +1,5 @@
 // OpenCSG - library for image-based CSG rendering for OpenGL
-// Copyright (C) 2002-2023, Florian Kirsch,
+// Copyright (C) 2002-2025, Florian Kirsch,
 // Hasso-Plattner-Institute at the University of Potsdam, Germany
 //
 // This library is free software; you can redistribute it and/or modify
@@ -17,7 +17,7 @@
 // 51 Franklin Street, Fifth Floor, Boston, MA 02110 - 1301 USA.
 
 //
-// settings.h 
+// settings.h
 //
 // global OpenCSG settings
 //
@@ -28,8 +28,14 @@
 #include "opencsgConfig.h"
 #include <opencsg.h>
 
+#include <string>
+
 namespace OpenCSG {
 
+    /// redeclared from opencsg.h
+    void setVertexShader(const std::string& vertexShader);
+    const char* getVertexShader();
+
     /// redeclared from opencsg.h
     void setOption(OptionType option, int newSetting);
     /// redeclared from opencsg.h
